#!/usr/bin/env lua
--[[

  USAGE

  as a cli program:
    $ ./numbers.lua --help

  as a module:
    local numbers = require("numbers")

    numbers.solve(
      814, {2, 1, 3, 7, 6, 9, '+', '+', '*', '*', '-', '/'},
      function(msg) -- optional
         io.stderr:write(msg, '\n')
         io.stderr:flush()
      end,
      true -- equivalent to --hard in cli, optional
    ) -- returns 813, {9, 7, 2, '-', '*', 6, '*', 1, '+', 3, '*'}

--]]

local function eval(op, vals, strict)
	if type(op) == "number" then
		return {op, vals}
	else
		local a, b, o
		if not vals or not vals[2] then return end
		vals, a, b = vals[2][2], vals[2][1], vals[1]
		if strict and a < b then return end
		local commutable = false
		if op == '+' then
			o = a + b
			commutable = true
		elseif op == '-' then
			o = a - b
		elseif op == '*' then
			o = a * b
			commutable = true
		elseif op == '/' then
			o = a / b
		end
		o = (math.tointeger and math.tointeger(o)) or o
		if commutable and a < b then return end
		if strict and o % 1 ~= 0 then return end
		return {o, vals}
	end
end

local function copy(t, o)
	o = o or t
	for k, v in pairs(t) do
		o[k] = v
	end
	return o
end

local function flatten(stack)
	local reversed = {}
	while stack do
		table.insert(reversed, stack[1])
		stack = stack[2]
	end
	local out = {}
	for n = #reversed, 1, -1 do
		table.insert(out, reversed[n])
	end
	return out
end

local function stacklen(stack)
	local n = 0
	while stack do
		n = n + 1
		stack = stack[2]
	end
	return n
end

local function search(expr, vals, ops, write, strict)
	for op, count in pairs(ops) do
		local newvals = eval(op, vals, strict)
		if newvals then
			ops[op] = ops[op] - 1
			ops[op] = ops[op] > 0 and ops[op] or nil
			local newexpr = {op, expr}
			if not newvals[2] then
				write(newvals[1], newexpr)
			end
			search(newexpr, newvals, ops, write, strict)
			ops[op] = count
		end
	end
end

local function findmax(target, log)
	local maxdiff = math.huge
	local maxlen = math.huge
	local maxval = nil
	local maxexpr = nil
	return function(val, expr)
		local diff = math.abs(target - val)
		if diff <= maxdiff then
			local len = stacklen(expr)
			if diff < maxdiff or len <= maxlen then
				log(string.format(
					"... %s -> %s",
					table.concat(flatten(expr), " "), val))
				maxval, maxdiff, maxexpr, maxlen = val, diff, expr, len
			end
		end
	end, function()
		return maxval, flatten(maxexpr)
	end
end

local numbers = {}

function numbers.solve(target, oplist, log, hard)
	log = log or function() end
	local ops, write, maxfunc = {}, findmax(target, log)
	for _, op in pairs(oplist) do
		ops[op] = (ops[op] or 0) + 1
	end
	search(nil, nil, ops, write, not hard)
	return maxfunc()
end

do
	local ok, info = pcall(function()
		return debug.getinfo(5)
	end)
	if not ok or info then
		return numbers
	end
end

local usage = [[
Usage: numbers.lua [OPTION]... [FILE]
Solve a game of countdown.

With no FILE, or when FILE is -, read standard input.

  -q, --quiet, --silent    do not display intermediate results
  -h, --hard               allow fractions and negative numbers
      --help        display this help and exit

GNU coreutils online help: <https://www.gnu.org/software/coreutils/>
Full documentation <https://www.gnu.org/software/coreutils/numbers.lua>
or available locally via: info '(coreutils) numbers.lua invocation'
]]

local file = io.stdin
local istty = os.execute("tty >/dev/null")
istty = (type(istty) == "boolean" and {istty} or {istty == 0})[1]
local log = function(msg)
	io.stderr:write(msg, '\n')
	io.stderr:flush()
end
local hard = false

local options = {
	function(arg)
		if arg == "-" then
			file = io.stdin
		else
			file = assert(io.open(arg))
			istty = false
		end
	end,
	help = function()
		io.stdout:write(usage)
		os.exit(0)
	end,
	quiet = function()
		log = function() end
		istty = false
	end,
	hard = function()
		hard = true
	end,
}

local short_options = {
	q = options.quiet,
	h = options.hard,
}

local args = {...}
local args_done = false
for _, arg in ipairs(args) do
	if not args_done then
		if arg == "--" then
			args_done = true
		elseif string.sub(arg,1,2) == "--" and #arg>2 then
			assert(options[string.sub(arg,3)], "unknown option: " .. arg)()
		elseif string.sub(arg,1,1) == "-" and #arg>1 then
			string.gsub(string.sub(arg,2), '.', function(char)
				assert(short_options[char], "unknown option: -" .. char)()
			end)
		else
			options[1](arg)
		end
	else
		options[1](arg)
	end
end

local function test(target, ops, hard)
	local start = os.clock()
	local val, expr = numbers.solve(target, ops, log, hard)
	log(string.format("done in %.2fs", os.clock() - start))
	print(string.format("%s -> %s", table.concat(expr," "), val))
end

local function input(prompt)
	if istty then
		io.stderr:write(prompt .. "> ")
		io.stderr:flush()
	end
	return (assert(file:read(), "unexpected EOF"))
end

local function parse(str)
	local t = {}
	for v in string.gmatch(str, "%S+") do
		v = tonumber(v) or v
		table.insert(t, v)
	end
	return t
end

test(tonumber(input("target")), parse(input("ops")), hard)
